/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef _GPXE_IPOIB_H
#define _GPXE_IPOIB_H

/** @file
 *
 * IP over Infiniband
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <gpxe/if_arp.h>
#include <gpxe/infiniband.h>

/** IPoIB packet length */
#define IPOIB_PKT_LEN 2048

/** IPoIB MAC address length */
#define IPOIB_ALEN 20

/** An IPoIB MAC address */
struct ipoib_mac {
	/** Queue pair number
	 *
	 * MSB must be zero; QPNs are only 24-bit.
	 */
	uint32_t qpn;
	/** Port GID */
	struct ib_gid gid;
} __attribute__ (( packed ));

/** IPoIB link-layer header length */
#define IPOIB_HLEN 4

/** IPoIB link-layer header */
struct ipoib_hdr {
	/** Network-layer protocol */
	uint16_t proto;
	/** Reserved, must be zero */
	uint16_t reserved;
} __attribute__ (( packed ));

/** GUID mask used for constructing eIPoIB Local Ethernet MAC address (LEMAC) */
#define IPOIB_GUID_MASK 0xe7

/** eIPoIB Remote Ethernet MAC address
 *
 * An eIPoIB REMAC address is an Ethernet-like (6 byte) link-layer
 * pseudo-address used to look up a full IPoIB link-layer address.
 */
struct ipoib_remac {
	/** Remote QPN
	 *
	 * Must be ORed with EIPOIB_QPN_LA so that eIPoIB REMAC
	 * addresses are considered as locally-assigned Ethernet MAC
	 * addreses.
	 */
	uint32_t qpn;
	/** Remote LID */
	uint16_t lid;
} __attribute__ (( packed ));

/** eIPoIB REMAC locally-assigned address indicator */
#define EIPOIB_QPN_LA 0x02000000UL

extern struct ll_protocol ipoib_protocol;

extern const char * ipoib_ntoa ( const void *ll_addr );

/**
 * Allocate IPoIB device
 *
 * @v priv_size		Size of driver private data
 * @ret netdev		Network device, or NULL
 */
static inline struct net_device * alloc_ipoibdev ( size_t priv_size ) {
	struct net_device *netdev;

	netdev = alloc_netdev ( priv_size );
	if ( netdev ) {
		netdev->ll_protocol = &ipoib_protocol;
		netdev->max_pkt_len = IPOIB_PKT_LEN;
	}
	return netdev;
}

extern void ipoib_link_state_changed ( struct ib_device *ibdev );
extern int ipoib_probe ( struct ib_device *ibdev );
extern void ipoib_remove ( struct ib_device *ibdev );

#endif /* _GPXE_IPOIB_H */
