/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */
#ifndef _GPXE_MALLOC_H
#define _GPXE_MALLOC_H

#include <stdint.h>

/** @file
 *
 * Dynamic memory allocation
 *
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

/*
 * Prototypes for the standard functions (malloc() et al) are in
 * stdlib.h.  Include <gpxe/malloc.h> only if you need the
 * non-standard functions, such as malloc_dma().
 *
 */
#include <stddef.h>
#include <stdlib.h>
#include <gpxe/tables.h>
#include <valgrind/memcheck.h>

#define NOWHERE ( ( void * ) ~( ( intptr_t ) 0 ) )

extern size_t freemem;

extern void * __malloc alloc_memblock ( size_t size, size_t align,
					size_t offset );
extern void free_memblock ( void *ptr, size_t size );
extern void mpopulate ( void *start, size_t len );
extern void mdumpfree ( void );

/**
 * Allocate memory for DMA
 *
 * @v size		Requested size
 * @v align		Physical alignment
 * @v offset		Offset from physical alignment
 * @ret ptr		Memory, or NULL
 *
 * Allocates physically-aligned memory for DMA.
 *
 * @c align must be a power of two.  @c size may not be zero.
 */
static inline void * __malloc malloc_dma_offset ( size_t size,
						  size_t phys_align,
						  size_t offset ) {
	void * ptr = alloc_memblock ( size, phys_align, offset );
	if (ptr == NOWHERE)
		ptr = NULL;
	if ( ptr && size )
		VALGRIND_MALLOCLIKE_BLOCK ( ptr, size, 0, 0 );
	return ptr;
}

#undef NOWHERE

/**
 * Allocate memory for DMA
 *
 * @v size		Requested size
 * @v align		Physical alignment
 * @ret ptr		Memory, or NULL
 *
 * Allocates physically-aligned memory for DMA.
 *
 * @c align must be a power of two.  @c size may not be zero.
 */
static inline void * __malloc malloc_dma ( size_t size, size_t phys_align ) {
	return malloc_dma_offset ( size, phys_align, 0 );
}

void free_dma ( void *ptr, size_t size );

/** A cache discarder */
struct cache_discarder {
	/**
	 * Discard some cached data
	 *
	 * @ret discarded	Number of cached items discarded
	 */
	unsigned int ( * discard ) ( void );
};

/** Cache discarder table */
#define CACHE_DISCARDERS __table ( struct cache_discarder, "cache_discarders" )

/** Declare a cache discarder */
#define __cache_discarder( cost ) __table_entry ( CACHE_DISCARDERS, cost )

/** @defgroup cache_cost Cache discarder costs
 *
 * @{
 */

#define CACHE_CHEAP	01	/**< Items with a low replacement cost */
#define CACHE_NORMAL	02	/**< Items with a normal replacement cost */
#define CACHE_EXPENSIVE	03	/**< Items with a high replacement cost */

/** @} */

#endif /* _GPXE_MALLOC_H */
