/*
 * SPDX-License-Identifier: GPL-2.0+
 *
 * Driver for Solarflare network controllers and boards
 * Copyright 2019 Solarflare Communications Inc.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation, incorporated herein by reference.
 *
 * You can also choose to distribute this program under the terms of
 * the Unmodified Binary Distribution Licence (as given in the file
 * COPYING.UBDL), provided that you have satisfied its requirements.
 */

FILE_LICENCE ( GPL2_OR_LATER_OR_UBDL );

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <byteswap.h>
#include <console.h>
#include <gpxe/keys.h>
#include <gpxe/vsprintf.h>
#include <gpxe/editstring.h>
#include <gpxe/net/sfc/pxe.h>
#include <readline/readline.h>

/** Boot servers setting */
struct setting boot_servers_setting __setting = {
	.name = "Boot servers",
	.description = "Types and addresses of servers to boot from",
	.tag = DHCP_PXE_BOOT_SERVERS,
	.type = &setting_type_string,
};

/** Boot menu setting */
struct setting boot_menu_setting __setting = {
	.name = "PXE Boot menu",
	.description = "PXE boot menu with boot servers types",
	.tag = DHCP_PXE_BOOT_MENU,
	.type = &setting_type_string,
};

/** Boot menu timeout setting */
struct setting boot_menu_prompt_setting __setting = {
	.name = "PXE menu timeout",
	.description = "PXE timeout and prompt for boot menu",
	.tag = DHCP_PXE_BOOT_MENU_PROMPT,
	.type = &setting_type_string,
};

/* Selected boot server type */
static uint16_t bootserver_type;

static uint8_t print_bootserver_choice ( char *boot_menu, int len_boot_menu,
                                         char *menu_prompt, int len_menu_prompt,
                                         char *prompt, uint16_t *types,
                                         int *types_num )
{
	int tmp_len;
	char *ptr = boot_menu;
	uint16_t boot_type;
	char desc[256];
	uint8_t desc_len;
	uint8_t timeout;

	*types_num = 0;

	printf( "\n********** Boot server type menu **********\n\n" );
	printf( "0) Local boot.\n" );
	for ( tmp_len = 0; tmp_len < len_boot_menu; )
	{
		( *types_num )++;
		boot_type = * ( ( uint16_t * ) ptr );
		types[ *types_num ] = boot_type;
		ptr += 2;
		desc_len = * ( ( uint8_t * ) ptr );
		ptr++;
		memcpy ( desc, ptr, desc_len );
		desc[ desc_len ] = '\0';
		tmp_len += 3 + desc_len;
		ptr += desc_len;
		printf("%d) %s Type = %d.\n", ( *types_num ), desc,
		       htons ( boot_type ) );
	}
	if ( len_menu_prompt > 0 )
	{
		timeout = * ( ( uint8_t * ) menu_prompt );
		ptr = menu_prompt + 1;
		memcpy ( prompt, ptr, len_menu_prompt - 1 );
		prompt[ ( len_menu_prompt - 1 ) ] = '\0';
		if ( timeout < 255 && timeout > 0 )
			printf( "%s  %3d : ", prompt, timeout );
		else
			printf( "%s  : ", prompt );
	}
	else
	{
		printf("Please, choose boot server type: ");
		return 255;
	}
	return timeout;
}

uint16_t pxe_get_bootserver_type ( char *boot_menu, int len_boot_menu,
				   char *menu_prompt, int len_menu_prompt )
{
	uint16_t            timeout = 0;
	int                 wait_count = 0;
	int                 key = 0;
	char                prompt[255];
	uint8_t             tmp_timeout;
	char               *line = NULL;
	unsigned long       choice = 256;
	uint16_t            types[16];
	int                 types_num = 0;

	if ( len_boot_menu > 0 )
	{
		timeout = print_bootserver_choice (
			boot_menu, len_boot_menu,
			menu_prompt, len_menu_prompt,
			prompt, types,
			&types_num );
		tmp_timeout = timeout;
		if ( timeout != 255 )
		{
			for ( wait_count = 0 ; wait_count < timeout * 10 ;
			      wait_count++ ) {
				if ( iskey() ) {
					key = getkey(1);
					break;
				}
				mdelay(100);
				if ( wait_count % 10 == 9 )
					printf( "\r%s  %3d : ", prompt,
					        --tmp_timeout );
			}
		}
		if ( timeout < 255 && timeout > 0 )
			printf( "\r%s  :     \b\b\b\b", prompt );
		if ( timeout == 255 || ( wait_count < timeout * 10 &&
		                         key !=  LF ) )
		{
			line = readline ( "" );
		}
		if ( line != NULL )
			choice = strtoul ( line, NULL, 0 );
		if ( ( int ) choice <= types_num )
			bootserver_type = types[ choice ];
		else if ( ( timeout == 0 && types_num > 0 ) ||
		          ( wait_count == timeout * 10 ) )
		{
			printf( " 1\n" );
			bootserver_type = types[ 1 ];
		}
	}
	free ( line );
	return ntohs ( bootserver_type );
}

int pxe_get_bootserver ( const char *ptr, uint8_t len, int seq_num,
                         struct in_addr *bs_addr)
{
	while ( len > 0 ) {
		if ( ( * ( ( uint16_t * ) ptr ) ) == bootserver_type ) {
			ptr += 2;
			if ( ( * ( ( uint8_t * ) ptr ) ) < seq_num )
				return ( -1 );
			ptr += 1 + 4 * ( seq_num - 1 );
			memcpy ( bs_addr, ptr, 4 );
			return 0;
		}
		len -= 3 + 4 * ( ( * ( ( uint8_t * ) ( ptr + 2 ) ) ) );
		ptr += 3 + 4 * ( ( * ( ( uint8_t * ) ( ptr + 2 ) ) ) );
	}
	return ( -1 );
}
